@@ -1,19 +1,23 @@
 # This Source Code Form is subject to the terms of the Mozilla Public
 # License, v. 2.0. If a copy of the MPL was not distributed with this
-# file, You can obtain one at http://mozilla.org/MPL/2.0/.
+# file, You can obtain one at https://mozilla.org/MPL/2.0/.
 
+from functools import partial
 from itertools import islice
 import json
 import re
 
 from glom import glom
 
-from . import siglists_utils
+from .siglists_utils import get_signature_list_content
 from .utils import (
     collapse,
     drop_bad_characters,
     drop_prefix_and_return_type,
+    get_crashing_thread,
+    override_values,
     parse_source_file,
+    strip_leading_zeros,
 )
 
 
@@ -28,8 +32,11 @@ def join_ignore_empty(delimiter, list_of_strings):
 class Rule:
     """Base class for Signature generation rules"""
 
-    def __init__(self):
-        self.name = self.__class__.__name__
+    @property
+    def name(self):
+        # NOTE(willkg): In order for this to work in siggen and in Socorro, we can't
+        # include the Python module path
+        return self.__class__.__name__
 
     def predicate(self, crash_data, result):
         """Whether or not to run this rule
@@ -58,51 +65,38 @@ class Rule:
         return True
 
 
-class SignatureTool:
-    """Stack walking signature generator base class
-
-    This defines the interface for classes that take a stack and generate a
-    signature from it.
-
-    Subclasses should implement the ``_do_generate`` method.
-
-    """
-
-    def generate(self, source_list, hang_type=0, crashed_thread=None, delimiter=" | "):
-        signature, notes, debug_notes = self._do_generate(
-            source_list, hang_type, crashed_thread, delimiter
-        )
-
-        return signature, notes, debug_notes
-
-    def _do_generate(self, source_list, hang_type, crashed_thread, delimiter):
-        raise NotImplementedError
-
-
-class CSignatureTool(SignatureTool):
+class CSignatureTool:
     """Generates signature from C/C++/Rust stacks.
 
-    This is the class for signature generation tools that work on breakpad
-    C/C++ stacks. It normalizes frames and then runs them through the siglists
-    to determine which frames should be part of the signature.
+    This is the class for signature generation tools that work on C/C++/Rust stacks. It
+    normalizes frames and then runs them through the siglists to determine which frames
+    should be part of the signature.
 
     """
 
     hang_prefixes = {-1: "hang", 1: "chromehang"}
 
-    def __init__(self):
+    def __init__(self, datadir=None):
+        """
+        :param datadir: the directory holding signature lists to use or ``None`` to
+            use included signature lists
+        """
         super().__init__()
 
-        self.irrelevant_signature_re = re.compile(
-            "|".join(siglists_utils.IRRELEVANT_SIGNATURE_RE)
+        self.datadir = datadir
+        if datadir is not None:
+            get_contents = partial(get_signature_list_content, source=datadir)
+        else:
+            get_contents = get_signature_list_content
+
+        self.irrelevant_signature_re = self.build_re(
+            get_contents("irrelevant_signature_re")
         )
-        self.prefix_signature_re = re.compile(
-            "|".join(siglists_utils.PREFIX_SIGNATURE_RE)
+        self.prefix_signature_re = self.build_re(get_contents("prefix_signature_re"))
+        self.signatures_with_line_numbers_re = self.build_re(
+            get_contents("signatures_with_line_numbers_re")
         )
-        self.signatures_with_line_numbers_re = re.compile(
-            "|".join(siglists_utils.SIGNATURES_WITH_LINE_NUMBERS_RE)
-        )
-        self.signature_sentinels = siglists_utils.SIGNATURE_SENTINELS
+        self.signature_sentinels = get_contents("signature_sentinels")
 
         self.collapse_arguments = True
 
@@ -111,6 +105,9 @@ class CSignatureTool(SignatureTool):
         self.fixup_hash = re.compile(r"::h[0-9a-fA-F]+$")
         self.fixup_lambda_numbers = re.compile(r"::\$_\d+::")
 
+    def build_re(self, lines):
+        return re.compile("|".join(lines))
+
     def normalize_rust_function(self, function, line):
         """Normalizes a single Rust frame with a function."""
         # Drop the prefix and return type if there is any
@@ -213,28 +210,16 @@ class CSignatureTool(SignatureTool):
         line=None,
         module_offset=None,
         offset=None,
-        normalized=None,
-        # eat any extra kwargs passed in
-        # fmt: off
-        **kwargs
-        # fmt: on
+        unloaded_modules=None,
     ):
         """Normalizes a single frame
 
-        Returns a structured conglomeration of the input parameters to serve as
-        a signature. The parameter names of this function reflect the exact
-        names of the fields from the jsonMDSW frame output. This allows this
-        function to be invoked by passing a frame as ``**a_frame``.
-
-        Sometimes, a frame may already have a normalized version cached. If
-        that exists, return it instead.
+        Returns a structured conglomeration of the input parameters to serve as a
+        signature. The parameter names of this function reflect the exact names of the
+        fields from the JSON MDSW frame output. This allows this function to be invoked
+        by passing a frame as ``**a_frame``.
 
         """
-        # If there's a cached normalized value, use that so we don't spend time
-        # figuring it out again
-        if normalized is not None:
-            return normalized
-
         if function:
             # If there's a filename and it ends in .rs, then normalize using
             # Rust rules
@@ -254,13 +239,86 @@ class CSignatureTool(SignatureTool):
             return f"{file}#{line}"
 
         # If there's an offset and no module/module_offset, use that
-        if not module and not module_offset and offset:
-            return f"@{offset}"
+        if not module and not module_offset:
+            if unloaded_modules:
+                # Use the first unloaded module and the offset or "0"
+                unloaded_module = unloaded_modules[0]
+                unloaded_module_module = unloaded_module.get("module")
+                unloaded_module_offsets = unloaded_module.get("offsets")
+                if unloaded_module_module and unloaded_module_offsets:
+                    return "(unloaded {}@{})".format(
+                        unloaded_module_module,
+                        strip_leading_zeros(unloaded_module_offsets[0]),
+                    )
+                elif unloaded_module_module:
+                    return "(unloaded {})".format(unloaded_module_module)
+
+            if offset:
+                return f"@{strip_leading_zeros(offset)}"
 
         # Return module/module_offset
-        return "{}@{}".format(module or "", module_offset)
+        return "{}@{}".format(module or "", strip_leading_zeros(module_offset))
 
-    def _do_generate(self, source_list, hang_type, crashed_thread, delimiter=" | "):
+    def frame_generator(self, frames):
+        """Yields frames one at a time from a list of frames
+
+        Note: this treats inlines as individual frames and yields those when they're
+        available.
+
+        :arg frames: list of frame dicts
+
+        :returns: generator of frame data for normalization
+
+        """
+        for frame in frames:
+            inlines = frame.get("inlines") or []
+            for inline in inlines:
+                yield {
+                    "module": frame.get("module"),
+                    "function": inline.get("function"),
+                    "file": inline.get("file"),
+                    "line": inline.get("line"),
+                    "module_offset": frame.get("module_offset"),
+                    "offset": frame.get("offset"),
+                }
+
+            yield {
+                "module": frame.get("module"),
+                "function": frame.get("function"),
+                "file": frame.get("file"),
+                "line": frame.get("line"),
+                "module_offset": frame.get("module_offset"),
+                "offset": frame.get("offset"),
+                # NOTE(gsvelto): unloaded modules can only appear in non-inlined frames
+                "unloaded_modules": frame.get("unloaded_modules"),
+            }
+
+    def create_frame_list(self, thread_data, make_modules_lower_case=False):
+        """Takes thread data and builds a list of frames
+
+        Note: this treats inlines as individual frames.
+
+        :arg thread_data: dict of thread data
+        :arg make_modules_lower_case: whether or not to lowercase module
+
+        :returns: list of normalized frames
+
+        """
+        normalized_frames = []
+        frames = thread_data.get("frames", [])
+        for frame in islice(self.frame_generator(frames), MAXIMUM_FRAMES_TO_CONSIDER):
+            # Bug #1544246. In Rust 1.34, the panic symbols are missing the module in
+            # symbols files. This fixes that by adding the module.
+            frame = fix_missing_module(frame)
+
+            if make_modules_lower_case and frame.get("module"):
+                frame["module"] = frame["module"].lower()
+
+            normalized_frame = self.normalize_frame(**frame)
+            normalized_frames.append(normalized_frame)
+        return normalized_frames
+
+    def generate(self, source_list, crashed_thread=None, delimiter=" | "):
         """Iterate over frames in the crash stack and generate a signature.
 
         First, we look for a sentinel frame and if we find one, we start with that.
@@ -278,6 +336,11 @@ class CSignatureTool(SignatureTool):
         notes = []
         debug_notes = []
 
+        if self.datadir is None:
+            debug_notes.append("using included signature lists")
+        else:
+            debug_notes.append(f"using signature lists from {self.datadir}")
+
         # Shorten source_list to the first sentinel found
         sentinel_locations = []
         for a_sentinel in self.signature_sentinels:
@@ -309,7 +372,10 @@ class CSignatureTool(SignatureTool):
                 continue
 
             # If the frame signature is a dll, remove the @xxxxx part.
-            if ".dll" in a_signature.lower():
+            if (
+                not a_signature.startswith("(unloaded")
+                and ".dll" in a_signature.lower()
+            ):
                 a_signature = a_signature.split("@")[0]
 
                 # If this trimmed DLL signature is the same as the previous frame's, skip it.
@@ -326,39 +392,29 @@ class CSignatureTool(SignatureTool):
 
             debug_notes.append(f'prefix; continue iterating: "{a_signature}"')
 
-        # Add a special marker for hang crash reports.
-        if hang_type:
-            debug_notes.append(
-                "hang_type {}: prepending {}".format(
-                    hang_type, self.hang_prefixes[hang_type]
-                )
-            )
-            new_signature_list.insert(0, self.hang_prefixes[hang_type])
-
         signature = delimiter.join(new_signature_list)
 
         # Handle empty signatures to explain why we failed generating them.
         if signature == "" or signature is None:
             if crashed_thread is None:
-                notes.append(
-                    "CSignatureTool: No signature could be created because we do not know which "
-                    "thread crashed"
-                )
+                notes.append("CSignatureTool: no crashing thread identified")
                 signature = "EMPTY: no crashing thread identified"
+
             else:
                 notes.append(
-                    "CSignatureTool: No proper signature could be created because no good data "
-                    "for the crashing thread ({}) was found".format(crashed_thread)
+                    f"CSignatureTool: no frame data for crashing thread ({crashed_thread})"
                 )
-                try:
+                if source_list:
+                    # The frames were probably all irrelevant, so pick the first one.
                     signature = source_list[0]
-                except IndexError:
+                else:
+                    # There wasn't any frame data to look at.
                     signature = "EMPTY: no frame data available"
 
         return signature, notes, debug_notes
 
 
-class JavaSignatureTool(SignatureTool):
+class JavaSignatureTool:
     """This is the signature generation class for Java signatures."""
 
     # The max length of a java exception description--if it's longer than this,
@@ -368,9 +424,7 @@ class JavaSignatureTool(SignatureTool):
     java_line_number_killer = re.compile(r"\.java\:\d+\)$")
     java_hex_addr_killer = re.compile(r"@[0-9a-f]{8}")
 
-    def _do_generate(
-        self, source, hang_type_unused=0, crashed_thread_unused=None, delimiter=": "
-    ):
+    def generate(self, source, delimiter=": "):
         if not isinstance(source, str):
             return (
                 "EMPTY: Java stack trace not in expected format",
@@ -515,40 +569,21 @@ class SignatureGenerationRule(Rule):
     3. Signature is generated by joining those frames with " | " between
        them.
 
-    This rule also generates the proto_signature which is the complete list
-    of normalized frames.
+    If it's a C/C++/Rust signature, this rule also adds to ``result.extra``:
+
+    * ``normalized_frames``: the list of normalized frames
+    * ``proto_signature``: a ``" | "`` delimited string of the normalized
+      frames
+
+    :param signature_list_dir: path to the directory with the signature lists to use or
+        ``None`` if you want to use the included ones
 
     """
 
-    def __init__(self):
+    def __init__(self, signature_list_dir=None):
         super().__init__()
         self.java_signature_tool = JavaSignatureTool()
-        self.c_signature_tool = CSignatureTool()
-
-    def _create_frame_list(
-        self, crashing_thread_mapping, make_modules_lower_case=False
-    ):
-        frame_signatures_list = []
-        for a_frame in islice(
-            crashing_thread_mapping.get("frames", []), MAXIMUM_FRAMES_TO_CONSIDER
-        ):
-            # Bug #1544246. In Rust 1.34, the panic symbols are missing the
-            # module in symbols files. This fixes that by adding the module.
-            a_frame = fix_missing_module(a_frame)
-
-            if make_modules_lower_case and "module" in a_frame:
-                a_frame["module"] = a_frame["module"].lower()
-
-            normalized_frame = self.c_signature_tool.normalize_frame(**a_frame)
-            a_frame["normalized"] = normalized_frame
-            frame_signatures_list.append(normalized_frame)
-        return frame_signatures_list
-
-    def _get_crashing_thread(self, crash_data):
-        try:
-            return int(crash_data.get("crashing_thread", 0))
-        except (TypeError, ValueError):
-            return 0
+        self.c_signature_tool = CSignatureTool(datadir=signature_list_dir)
 
     def action(self, crash_data, result):
         # If this is a Java crash, then generate a Java signature
@@ -566,18 +601,12 @@ class SignatureGenerationRule(Rule):
 
         result.debug(self.name, "using CSignatureTool")
         try:
-            # First, we need to figure out which thread to look at. If it's a
-            # chrome hang (1), then use thread 0. Otherwise, use the crashing
-            # thread specified in the crash data.
-            if crash_data.get("hang_type", None) == 1:
-                crashing_thread = 0
-            else:
-                crashing_thread = self._get_crashing_thread(crash_data)
+            crashing_thread = get_crashing_thread(crash_data)
 
             # If we have a thread to look at, pull the frames for that.
             # Otherwise we don't have frames to use.
             if crashing_thread is not None:
-                signature_list = self._create_frame_list(
+                signature_list = self.c_signature_tool.create_frame_list(
                     glom(crash_data, "threads.%d" % crashing_thread, default={}),
                     crash_data.get("os") == "Windows NT",
                 )
@@ -591,12 +620,15 @@ class SignatureGenerationRule(Rule):
 
         signature, notes, debug_notes = self.c_signature_tool.generate(
             signature_list,
-            crash_data.get("hang_type"),
-            crash_data.get("crashing_thread"),
+            crashing_thread,
         )
 
         if signature_list:
+            # Capture the proto_signature which gets indexed in Elasticsearch
+            # and is searchable
             result.extra["proto_signature"] = " | ".join(signature_list)
+            # Capture the normalized_frames for users of signature generation
+            result.extra["normalized_frames"] = signature_list
         for note in notes:
             result.info(self.name, note)
         for note in debug_notes:
@@ -607,6 +639,32 @@ class SignatureGenerationRule(Rule):
         return True
 
 
+class StackOverflowSignature(Rule):
+    """Prepends ``stackoverflow``
+
+    See bug #1796389.
+
+    """
+
+    # These reason values indicate a stackoverflow
+    stackoverflow_reason = [
+        "EXCEPTION_STACK_OVERFLOW",
+    ]
+
+    def predicate(self, crash_data, result):
+        # Check the reason to see if it's one of a few values that indicate a
+        # stackoverflow
+        reason = crash_data.get("reason", None)
+        if reason in self.stackoverflow_reason:
+            return True
+
+        return False
+
+    def action(self, crash_data, result):
+        result.set_signature(self.name, f"stackoverflow | {result.signature}")
+        return True
+
+
 class OOMSignature(Rule):
     """Prepends ``OOM | <size>`` to signatures for OOM crashes.
 
@@ -614,6 +672,7 @@ class OOMSignature(Rule):
 
     """
 
+    # Fragments that are in the signature that indicate this is an OOM
     signature_fragments = (
         "NS_ABORT_OOM",
         "mozalloc_handle_oom",
@@ -622,6 +681,18 @@ class OOMSignature(Rule):
         "alloc::oom::oom",
     )
 
+    # These reason values indicate an OOM
+    oom_reason = [
+        "STATUS_FATAL_MEMORY_EXHAUSTION",
+        "STATUS_NO_MEMORY",
+    ]
+
+    # These last_error_value values indicate an OOM
+    oom_last_error_value = [
+        "ERROR_COMMITMENT_LIMIT",
+        "ERROR_NOT_ENOUGH_MEMORY",
+    ]
+
     def predicate(self, crash_data, result):
         if crash_data.get("oom_allocation_size"):
             return True
@@ -634,6 +705,20 @@ class OOMSignature(Rule):
             if a_signature_fragment in signature:
                 return True
 
+        # Check the last_error_value of the crashing thread to see if it's
+        # ERROR_COMMITMENT_LIMIT which indicates this is an OOM
+        crashing_thread = get_crashing_thread(crash_data)
+        thread = glom(crash_data, "threads.%d" % crashing_thread, default={})
+        if thread:
+            last_error_value = thread.get("last_error_value", "")
+            if last_error_value in self.oom_last_error_value:
+                return True
+
+        # Check the reason to see if it's one of a few values that indicate an OOM
+        reason = crash_data.get("reason", None)
+        if reason in self.oom_reason:
+            return True
+
         return False
 
     def action(self, crash_data, result):
@@ -650,6 +735,34 @@ class OOMSignature(Rule):
         return True
 
 
+class BadHardware(Rule):
+    """Prepends ``bad hardware`` to signatures that are from bad hardware.
+
+    See bug #1733904.
+
+    """
+
+    # These reason value substrings indicate an error caused by bad hardware
+    bad_hardware_reason = [
+        "STATUS_DEVICE_DATA_ERROR",
+    ]
+
+    def predicate(self, crash_data, result):
+        reason = crash_data.get("reason", None)
+        if not reason:
+            return False
+
+        for possibility in self.bad_hardware_reason:
+            if possibility in reason:
+                return True
+
+        return False
+
+    def action(self, crash_data, result):
+        result.set_signature(self.name, f"bad hardware | {result.signature}")
+        return True
+
+
 class AbortSignature(Rule):
     """Prepends abort message to signature.
 
@@ -763,30 +876,45 @@ class StackwalkerErrorSignatureRule(Rule):
         return True
 
 
-class SignatureRunWatchDog(SignatureGenerationRule):
+class SignatureRunWatchDog(Rule):
     """Prepends "shutdownhang" to signature for shutdown hang crashes."""
 
+    def __init__(self):
+        super().__init__()
+        self.signature_generation_rule = SignatureGenerationRule()
+
     def predicate(self, crash_data, result):
         return "RunWatchdog" in result.signature
 
-    def _get_crashing_thread(self, crash_data):
-        # Always use thread 0 in this case, because that's the thread that
-        # was hanging when the software was artificially crashed.
-        return 0
-
     def action(self, crash_data, result):
-        # For shutdownhang crashes, we need to use thread 0 instead of the
-        # crashing thread. The reason is because those crashes happen
-        # artificially when thread 0 gets stuck. So whatever the crashing
-        # thread is, we don't care about it and only want to know what was
-        # happening in thread 0 when it got stuck.
-        ret = super().action(crash_data, result)
+        # For shutdownhang crashes, we need to use thread 0 instead of the crashing
+        # thread. The reason is because those crashes happen artificially when thread 0
+        # gets stuck. So whatever the crashing thread is, we don't care about it and
+        # only want to know what was happening in thread 0 when it got stuck.
+        with override_values(crash_data, {"crashing_thread": 0}) as crash_data:
+            ret = self.signature_generation_rule.action(crash_data, result)
+
         result.set_signature(self.name, f"shutdownhang | {result.signature}")
         return ret
 
 
 class SignatureShutdownTimeout(Rule):
-    """Replaces signature with async_shutdown_timeout message."""
+    """Replaces signature with async_shutdown_timeout message.
+
+    This supports AsyncShutdownTimeout annotation values with the following structure::
+
+        {
+            "phase": <str>,
+            "conditions": [
+                {
+                    "name": <str>,
+                    ...
+                }
+            ]
+        }
+
+
+    """
 
     def predicate(self, crash_data, result):
         return bool(crash_data.get("async_shutdown_timeout"))
@@ -795,14 +923,18 @@ class SignatureShutdownTimeout(Rule):
         parts = ["AsyncShutdownTimeout"]
         try:
             shutdown_data = json.loads(crash_data["async_shutdown_timeout"])
-            parts.append(shutdown_data["phase"])
+            if isinstance(shutdown_data.get("phase"), str):
+                parts.append(shutdown_data["phase"])
+            else:
+                parts.append("(unknown)")
+
             conditions = [
-                # NOTE(willkg): The AsyncShutdownTimeout notation condition can either be a string
-                # that looks like a "name" or a dict with a "name" in it.
+                # NOTE(willkg): The AsyncShutdownTimeout notation condition can either
+                # be a string that looks like a "name" or a dict with a "name" in it.
                 #
                 # This handles both variations.
                 c["name"] if isinstance(c, dict) else c
-                for c in shutdown_data["conditions"]
+                for c in shutdown_data.get("conditions") or []
             ]
             if conditions:
                 conditions.sort()
@@ -823,30 +955,10 @@ class SignatureShutdownTimeout(Rule):
         return True
 
 
-class SignatureJitCategory(Rule):
-    """Replaces signature with JIT classification."""
-
-    def predicate(self, crash_data, result):
-        return bool(crash_data.get("jit_category"))
-
-    def action(self, crash_data, result):
-        result.info(
-            self.name,
-            'Signature replaced with a JIT Crash Category, was: "%s"',
-            result.signature,
-        )
-        result.set_signature(self.name, "jit | {}".format(crash_data["jit_category"]))
-        return True
-
-
 class SignatureIPCChannelError(Rule):
-    """Either stomp on or prepend signature for IPCError
+    """Stomps on signature with shutdownkill signature
 
-    If the IPCError is a ShutDownKill, then this prepends the signature with
-    "IPCError-browser | ShutDownKill".
-
-    Otherwise it stomps on the signature with "IPCError-browser/content" and the error
-    message.
+    Either "IPCError-browser | ShutDownKill" or "IPCError-content | ShutDownKill".
 
     """
 
@@ -854,19 +966,13 @@ class SignatureIPCChannelError(Rule):
         return bool(crash_data.get("ipc_channel_error"))
 
     def action(self, crash_data, result):
-        if crash_data.get("additional_minidumps") == "browser":
+        minidumps = crash_data.get("additional_minidumps") or []
+        if "upload_file_minidump_browser" in minidumps:
             new_sig = "IPCError-browser | {}"
         else:
             new_sig = "IPCError-content | {}"
         new_sig = new_sig.format(crash_data["ipc_channel_error"][:100])
-
-        if crash_data["ipc_channel_error"] == "ShutDownKill":
-            # If it's a ShutDownKill, append the rest of the signature
-            result.info(self.name, "IPC Channel Error prepended")
-            new_sig = f"{new_sig} | {result.signature}"
-        else:
-            result.info(self.name, "IPC Channel Error stomped on signature")
-
+        result.info(self.name, "IPC Channel Error stomped on signature")
         result.set_signature(self.name, new_sig)
         return True
 
