@@ -1,21 +1,21 @@
 # This Source Code Form is subject to the terms of the Mozilla Public
 # License, v. 2.0. If a copy of the MPL was not distributed with this
-# file, You can obtain one at http://mozilla.org/MPL/2.0/.
-
+# file, You can obtain one at https://mozilla.org/MPL/2.0/.
 
+import copy
 import importlib
 import json
 import re
-from unittest import mock
 
 import pytest
 
+
 # NOTE(willkg): We do this so that we can extract signature generation into its
 # own namespace as an external library. This allows the tests to run if it's in
 # "siggen" or "socorro.signature".
-base_module = ".".join(__name__.split(".")[:-2])
-rules = importlib.import_module(base_module + ".rules")
-generator = importlib.import_module(base_module + ".generator")
+BASE_MODULE = ".".join(__name__.split(".")[:-2])
+rules = importlib.import_module(BASE_MODULE + ".rules")
+generator = importlib.import_module(BASE_MODULE + ".generator")
 
 
 @pytest.mark.parametrize(
@@ -49,13 +49,12 @@ class TestCSignatureTool:
         td=[r"foo32\.dll.*"],
         ss=("sentinel", ("sentinel2", lambda x: "ff" in x)),
     ):
-        with mock.patch(base_module + ".rules.siglists_utils") as mocked_siglists:
-            mocked_siglists.IRRELEVANT_SIGNATURE_RE = ig
-            mocked_siglists.PREFIX_SIGNATURE_RE = pr
-            mocked_siglists.SIGNATURES_WITH_LINE_NUMBERS_RE = si
-            mocked_siglists.TRIM_DLL_SIGNATURE_RE = td
-            mocked_siglists.SIGNATURE_SENTINELS = ss
-            return rules.CSignatureTool()
+        tool = rules.CSignatureTool()
+        tool.irrelevant_signature_re = tool.build_re(ig)
+        tool.prefix_signature_re = tool.build_re(pr)
+        tool.signatures_with_line_numbers_re = tool.build_re(si)
+        tool.signature_sentinels = ss
+        return tool
 
     def test_c_config_tool_init(self):
         """test_C_config_tool_init: constructor test"""
@@ -77,19 +76,185 @@ class TestCSignatureTool:
         assert fixup_space.pattern == s.fixup_space.pattern
         assert fixup_comma.pattern == s.fixup_comma.pattern
 
-    def test_normalize_frame(self):
-        """test_normalize: bunch of variations"""
+    def test_create_frame_list(self):
+        thread_0 = {
+            "frames": [
+                {
+                    "frame": 0,
+                    "function": "NtWaitForMultipleObjects",
+                    "function_offset": "0x15",
+                    "module": "ntdll.dll",
+                    "module_offset": "0x2015d",
+                    "offset": "0x77ad015d",
+                    "trust": "context",
+                },
+                {
+                    "frame": 1,
+                    "function": "WaitForMultipleObjectsEx",
+                    "function_offset": "0xff",
+                    "module": "KERNELBASE.dll",
+                    "module_offset": "0x115f6",
+                    "offset": "0x775e15f6",
+                    "trust": "cfi",
+                },
+                {
+                    "frame": 2,
+                    "function": "WaitForMultipleObjectsExImplementation",
+                    "function_offset": "0x8d",
+                    "module": "kernel32.dll",
+                    "module_offset": "0x119f7",
+                    "offset": "0x766119f7",
+                    "trust": "cfi",
+                },
+                {
+                    "frame": 3,
+                    "function": "RealMsgWaitForMultipleObjectsEx",
+                    "function_offset": "0xe1",
+                    "module": "user32.dll",
+                    "module_offset": "0x20869",
+                    "offset": "0x77370869",
+                    "trust": "cfi",
+                },
+                {
+                    "frame": 4,
+                    "function": "MsgWaitForMultipleObjects",
+                    "function_offset": "0x1e",
+                    "module": "user32.dll",
+                    "module_offset": "0x20b68",
+                    "offset": "0x77370b68",
+                    "trust": "cfi",
+                },
+                {
+                    "file": "F_766591945_______________________________________",
+                    "frame": 5,
+                    "function": "F_1428703866________________________________",
+                    "function_offset": "0xc1",
+                    "line": 203,
+                    "module": "NPSWF32_14_0_0_125.dll",
+                    "module_offset": "0x35bf21",
+                    "offset": "0x5e39bf21",
+                    "trust": "cfi",
+                },
+            ],
+            "threads_index": 0,
+            "frame_count": 32,
+        }
+
         s = self.setup_config_c_sig_tool()
-        a = [
-            (("module", "", "source/", "23", "0xFFF"), "source#23"),
-            (("module", "", "source\\", "23", "0xFFF"), "source#23"),
-            (("module", "", "/a/b/c/source", "23", "0xFFF"), "source#23"),
-            (("module", "", "\\a\\b\\c\\source", "23", "0xFFF"), "source#23"),
-            (("module", "", "\\a\\b\\c\\source", "23", "0xFFF"), "source#23"),
-            (("module", "", "\\a\\b\\c\\source", "", "0xFFF"), "module@0xFFF"),
-            (("module", "", "", "23", "0xFFF"), "module@0xFFF"),
-            (("module", "", "", "", "0xFFF"), "module@0xFFF"),
-            ((None, "", "", "", "0xFFF"), "@0xFFF"),
+        frame_signatures_list = s.create_frame_list(thread_0)
+        expected = [
+            "NtWaitForMultipleObjects",
+            "WaitForMultipleObjectsEx",
+            "WaitForMultipleObjectsExImplementation",
+            "RealMsgWaitForMultipleObjectsEx",
+            "MsgWaitForMultipleObjects",
+            "F_1428703866________________________________",
+        ]
+        assert frame_signatures_list == expected
+
+    def test_create_frame_list_with_inlines(self):
+        thread_0 = {
+            "frames": [
+                {
+                    "frame": 0,
+                    "function": "NtWaitForMultipleObjects",
+                    "function_offset": "0x15",
+                    "module": "ntdll.dll",
+                    "module_offset": "0x2015d",
+                    "offset": "0x77ad015d",
+                    "trust": "context",
+                    "inlines": None,
+                },
+                {
+                    "frame": 1,
+                    "function": "WaitForMultipleObjectsEx",
+                    "function_offset": "0xff",
+                    "module": "KERNELBASE.dll",
+                    "module_offset": "0x115f6",
+                    "offset": "0x775e15f6",
+                    "trust": "cfi",
+                    "inlines": [
+                        {
+                            "function": "inline_a",
+                            "file": "some_file.rs",
+                            "line": 59,
+                        },
+                        {
+                            "function": "inline_b",
+                            "file": "some_file.rs",
+                            "line": 299,
+                        },
+                    ],
+                },
+                {
+                    "frame": 2,
+                    "function": "WaitForMultipleObjectsExImplementation",
+                    "function_offset": "0x8d",
+                    "module": "kernel32.dll",
+                    "module_offset": "0x119f7",
+                    "offset": "0x766119f7",
+                    "trust": "cfi",
+                },
+            ],
+            "threads_index": 0,
+            "frame_count": 3,
+        }
+
+        s = self.setup_config_c_sig_tool()
+        frame_signatures_list = s.create_frame_list(thread_0)
+        expected = [
+            "NtWaitForMultipleObjects",
+            "inline_a",
+            "inline_b",
+            "WaitForMultipleObjectsEx",
+            "WaitForMultipleObjectsExImplementation",
+        ]
+        assert frame_signatures_list == expected
+
+    def test_no_mutation(self):
+        s = self.setup_config_c_sig_tool()
+        frames = {
+            "frames": [
+                {
+                    "frame": 0,
+                    "function": "NtWaitForMultipleObjects",
+                    "function_offset": "0x15",
+                    "module": "ntdll.dll",
+                    "module_offset": "0x2015d",
+                    "offset": "0x77ad015d",
+                    "trust": "context",
+                },
+                {
+                    "frame": 1,
+                    "function": "WaitForMultipleObjectsEx",
+                    "function_offset": "0xff",
+                    "module": "KERNELBASE.dll",
+                    "module_offset": "0x115f6",
+                    "offset": "0x775e15f6",
+                    "trust": "cfi",
+                },
+            ]
+        }
+        frames_copy = copy.deepcopy(frames)
+        s = self.setup_config_c_sig_tool()
+        s.create_frame_list(frames_copy)
+
+        # If they're the same, then there was no mutation
+        assert frames == frames_copy
+
+    @pytest.mark.parametrize(
+        "args, expected",
+        [
+            # module, function, file, line, module_offset, offset, unloaded_modules
+            (("module", "", "source/", "23", "0xfff"), "source#23"),
+            (("module", "", "source\\", "23", "0xfff"), "source#23"),
+            (("module", "", "/a/b/c/source", "23", "0xfff"), "source#23"),
+            (("module", "", "\\a\\b\\c\\source", "23", "0xfff"), "source#23"),
+            (("module", "", "\\a\\b\\c\\source", "23", "0xfff"), "source#23"),
+            (("module", "", "\\a\\b\\c\\source", "", "0xfff"), "module@0xfff"),
+            (("module", "", "", "23", "0xfff"), "module@0xfff"),
+            (("module", "", "", "", "0xfff"), "module@0xfff"),
+            ((None, "", "", "", "0xfff"), "@0xfff"),
             # Make sure frame normalization uses the right function: normalize
             # Rust frame (has a Rust fingerprint)
             (
@@ -98,14 +263,42 @@ class TestCSignatureTool:
                     "expect_failed::h7f635057bfba806a",
                     "hg:hg.mozilla.org/a/b:servio/wrapper.rs:44444444444",
                     "23",
-                    "0xFFF",
+                    "0xfff",
                 ),
                 "expect_failed",
             ),
-        ]
-        for args, e in a:
-            r = s.normalize_frame(*args)
-            assert e == r
+            # Handle normalization with unloaded modules
+            (
+                (
+                    "",
+                    "",
+                    "",
+                    "",
+                    "",
+                    "",
+                    [{"module": "unmod"}],
+                ),
+                "(unloaded unmod)",
+            ),
+            (
+                (
+                    "",
+                    "",
+                    "",
+                    "",
+                    "",
+                    "",
+                    [{"module": "unmod", "offsets": ["0x0000000000005387"]}],
+                ),
+                "(unloaded unmod@0x5387)",
+            ),
+        ],
+    )
+    def test_normalize_frame(self, args, expected):
+        """test_normalize: bunch of variations"""
+        s = self.setup_config_c_sig_tool()
+        ret = s.normalize_frame(*args)
+        assert expected == ret
 
     @pytest.mark.parametrize(
         "function, line, expected",
@@ -220,7 +413,7 @@ class TestCSignatureTool:
 
     def test_generate_1(self):
         """test_generate_1: simple"""
-        s = self.setup_config_c_sig_tool(["a", "b", "c"], ["d", "e", "f"])
+        s = self.setup_config_c_sig_tool(ig=["a", "b", "c"], pr=["d", "e", "f"])
         a = list("abcdefghijklmnopqrstuvwxyz")
         sig, notes, debug_notes = s.generate(a)
         assert sig == "d | e | f | g"
@@ -229,28 +422,9 @@ class TestCSignatureTool:
         sig, notes, debug_notes = s.generate(a)
         assert sig == "d | e | f | g"
 
-    def test_generate_2(self):
-        """test_generate_2: hang"""
-        s = self.setup_config_c_sig_tool(["a", "b", "c"], ["d", "e", "f"])
-        a = list("abcdefghijklmnopqrstuvwxyz")
-        sig, notes, debug_notes = s.generate(a, hang_type=-1)
-        assert sig == "hang | d | e | f | g"
-
-        a = list("abcdaeafagahijklmnopqrstuvwxyz")
-        sig, notes, debug_notes = s.generate(a, hang_type=-1)
-        assert sig == "hang | d | e | f | g"
-
-        a = list("abcdaeafagahijklmnopqrstuvwxyz")
-        sig, notes, debug_notes = s.generate(a, hang_type=0)
-        assert sig == "d | e | f | g"
-
-        a = list("abcdaeafagahijklmnopqrstuvwxyz")
-        sig, notes, debug_notes = s.generate(a, hang_type=1)
-        assert sig == "chromehang | d | e | f | g"
-
     def test_generate_2a(self):
         """test_generate_2a: way too long"""
-        s = self.setup_config_c_sig_tool(["a", "b", "c"], ["d", "e", "f"])
+        s = self.setup_config_c_sig_tool(ig=["a", "b", "c"], pr=["d", "e", "f"])
         a = list("abcdefghijklmnopqrstuvwxyz")
         a[3] = a[3] * 70
         a[4] = a[4] * 70
@@ -265,44 +439,39 @@ class TestCSignatureTool:
         )
         sig, notes, debug_notes = s.generate(a)
         assert sig == expected
-        expected = (
-            "hang | "
-            "dddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddd | "
-            "eeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeee | "
-            "ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff | "
-            "gggggggggggggggggggggggggggggggggggggggggggggggggggggggggggggggggggggg"
-        )
-        sig, notes, debug_notes = s.generate(a, hang_type=-1)
-        assert sig == expected
 
     def test_generate_3(self):
         """test_generate_3: simple sentinel"""
-        s = self.setup_config_c_sig_tool(["a", "b", "c"], ["d", "e", "f"])
+        s = self.setup_config_c_sig_tool(ig=["a", "b", "c"], pr=["d", "e", "f"])
         a = list("abcdefghabcfaeabdijklmnopqrstuvwxyz")
         a[7] = "sentinel"
         sig, notes, debug_notes = s.generate(a)
         assert sig == "sentinel"
 
-        s = self.setup_config_c_sig_tool(["a", "b", "c", "sentinel"], ["d", "e", "f"])
+        s = self.setup_config_c_sig_tool(
+            ig=["a", "b", "c", "sentinel"], pr=["d", "e", "f"]
+        )
         sig, notes, debug_notes = s.generate(a)
         assert sig == "f | e | d | i"
 
     def test_generate_4(self):
         """test_generate_4: tuple sentinel"""
-        s = self.setup_config_c_sig_tool(["a", "b", "c"], ["d", "e", "f"])
+        s = self.setup_config_c_sig_tool(ig=["a", "b", "c"], pr=["d", "e", "f"])
         a = list("abcdefghabcfaeabdijklmnopqrstuvwxyz")
         a[7] = "sentinel2"
         sig, notes, debug_notes = s.generate(a)
         assert sig == "d | e | f | g"
 
-        s = self.setup_config_c_sig_tool(["a", "b", "c"], ["d", "e", "f"])
+        s = self.setup_config_c_sig_tool(ig=["a", "b", "c"], pr=["d", "e", "f"])
         a = list("abcdefghabcfaeabdijklmnopqrstuvwxyz")
         a[7] = "sentinel2"
         a[22] = "ff"
         sig, notes, debug_notes = s.generate(a)
         assert sig == "sentinel2"
 
-        s = self.setup_config_c_sig_tool(["a", "b", "c", "sentinel2"], ["d", "e", "f"])
+        s = self.setup_config_c_sig_tool(
+            ig=["a", "b", "c", "sentinel2"], pr=["d", "e", "f"]
+        )
         a = list("abcdefghabcfaeabdijklmnopqrstuvwxyz")
         a[7] = "sentinel2"
         a[22] = "ff"
@@ -310,19 +479,19 @@ class TestCSignatureTool:
         assert sig == "f | e | d | i"
 
     def test_generate_with_merged_dll(self):
-        sig_tool = self.setup_config_c_sig_tool(["a", "b", "c"], ["d", "e", "f"])
-        source_list = (
+        sig_tool = self.setup_config_c_sig_tool(ig=["a", "b", "c"], pr=["d", "e", "f"])
+        source_list = [
             "a",
             "d",
             "foo32.dll@0x231423",
             "foo32.dll",
             "foo32.dll@0x42",
             "g",
-        )
+        ]
         sig, notes, debug_notes = sig_tool.generate(source_list)
         assert sig == "d | foo32.dll | g"
 
-        source_list = ("foo32.dll", "foo32.dll@0x231423", "g")
+        source_list = ["foo32.dll", "foo32.dll@0x231423", "g"]
         sig, notes, debug_notes = sig_tool.generate(source_list)
         assert sig == "foo32.dll | g"
 
@@ -547,7 +716,7 @@ java.lang.IllegalArgumentException: Receiver not registered: org.mozilla.gecko.G
 
 #  rules testing section
 
-frames_from_json_dump = {
+FRAMES_FROM_JSON_DUMP = {
     "frames": [
         {
             "frame": 0,
@@ -666,7 +835,7 @@ frames_from_json_dump = {
     "frame_count": 32,
 }
 
-frames_from_json_dump_with_templates = {
+FRAMES_FROM_JSON_DUMP_WITH_TEMPLATES = {
     "frames": [
         {
             "frame": 0,
@@ -782,10 +951,9 @@ frames_from_json_dump_with_templates = {
         },
     ],
     "threads_index": 0,
-    "total_frames": 32,
 }
 
-frames_from_json_dump_with_templates_and_special_case = {
+FRAMES_FROM_JSON_DUMP_WITH_TEMPLATES_AND_SPECIAL_CASE = {
     "frames": [
         {
             "frame": 0,
@@ -901,30 +1069,10 @@ frames_from_json_dump_with_templates_and_special_case = {
         },
     ],
     "threads_index": 0,
-    "total_frames": 32,
 }
 
 
-class TestSignatureGeneration:
-    def test_create_frame_list(self):
-        sgr = rules.SignatureGenerationRule()
-        frame_signatures_list = sgr._create_frame_list(frames_from_json_dump)
-        expected = [
-            "NtWaitForMultipleObjects",
-            "WaitForMultipleObjectsEx",
-            "WaitForMultipleObjectsExImplementation",
-            "RealMsgWaitForMultipleObjectsEx",
-            "MsgWaitForMultipleObjects",
-            "F_1152915508__________________________________",
-            "F2166389______________________________________",
-            "F_917831355___________________________________",
-            "F1315696776________________________________",
-            "F_1428703866________________________________",
-        ]
-        assert frame_signatures_list == expected
-        assert "normalized" in frames_from_json_dump["frames"][0]
-        assert frames_from_json_dump["frames"][0]["normalized"] == expected[0]
-
+class TestSignatureGenerationRule:
     def test_java_stack_trace(self):
         sgr = rules.SignatureGenerationRule()
 
@@ -943,6 +1091,7 @@ class TestSignatureGeneration:
         expected = "SomeJavaException: at org.mozilla.lars.myInvention(larsFile.java)"
         assert result.signature == expected
         assert "proto_signature" not in result.extra
+        assert "normalized_frames" not in result.extra
         assert result.notes == [
             "SignatureGenerationRule: JavaSignatureTool: dropped Java exception description due to length"  # noqa
         ]
@@ -950,7 +1099,7 @@ class TestSignatureGeneration:
     def test_c_stack_trace(self):
         sgr = rules.SignatureGenerationRule()
 
-        crash_data = {"os": "Windows NT", "threads": [frames_from_json_dump]}
+        crash_data = {"os": "Windows NT", "threads": [FRAMES_FROM_JSON_DUMP]}
         result = generator.Result()
 
         # the call to be tested
@@ -961,17 +1110,22 @@ class TestSignatureGeneration:
         )
         assert result.signature == expected
 
-        expected = (
-            "NtWaitForMultipleObjects | WaitForMultipleObjectsEx | "
-            "WaitForMultipleObjectsExImplementation | "
-            "RealMsgWaitForMultipleObjectsEx | MsgWaitForMultipleObjects | "
-            "F_1152915508__________________________________ | "
-            "F2166389______________________________________ | "
-            "F_917831355___________________________________ | "
-            "F1315696776________________________________ | "
-            "F_1428703866________________________________"
-        )
-        assert result.extra["proto_signature"] == expected
+        expected_normalized_frames = [
+            "NtWaitForMultipleObjects",
+            "WaitForMultipleObjectsEx",
+            "WaitForMultipleObjectsExImplementation",
+            "RealMsgWaitForMultipleObjectsEx",
+            "MsgWaitForMultipleObjects",
+            "F_1152915508__________________________________",
+            "F2166389______________________________________",
+            "F_917831355___________________________________",
+            "F1315696776________________________________",
+            "F_1428703866________________________________",
+        ]
+        expected_proto_signature = " | ".join(expected_normalized_frames)
+
+        assert result.extra["proto_signature"] == expected_proto_signature
+        assert result.extra["normalized_frames"] == expected_normalized_frames
         assert result.notes == []
 
     def test_action_2_with_templates(self):
@@ -980,7 +1134,7 @@ class TestSignatureGeneration:
         crash_data = {
             "os": "Windows NT",
             "crashing_thread": 0,
-            "threads": [frames_from_json_dump_with_templates],
+            "threads": [FRAMES_FROM_JSON_DUMP_WITH_TEMPLATES],
         }
         result = generator.Result()
 
@@ -988,18 +1142,21 @@ class TestSignatureGeneration:
         assert sgr.action(crash_data, result) is True
 
         assert result.signature == "Alpha<T>::Echo<T>"
-        expected = (
-            "NtWaitForMultipleObjects | Alpha<T>::Echo<T> | "
-            "WaitForMultipleObjectsExImplementation | "
-            "RealMsgWaitForMultipleObjectsEx | "
-            "MsgWaitForMultipleObjects | "
-            "F_1152915508__________________________________ | "
-            "F2166389______________________________________ | "
-            "F_917831355___________________________________ | "
-            "F1315696776________________________________ | "
-            "F_1428703866________________________________"
-        )
-        assert result.extra["proto_signature"] == expected
+        expected_normalized_frames = [
+            "NtWaitForMultipleObjects",
+            "Alpha<T>::Echo<T>",
+            "WaitForMultipleObjectsExImplementation",
+            "RealMsgWaitForMultipleObjectsEx",
+            "MsgWaitForMultipleObjects",
+            "F_1152915508__________________________________",
+            "F2166389______________________________________",
+            "F_917831355___________________________________",
+            "F1315696776________________________________",
+            "F_1428703866________________________________",
+        ]
+        expected_proto_signature = " | ".join(expected_normalized_frames)
+        assert result.extra["proto_signature"] == expected_proto_signature
+        assert result.extra["normalized_frames"] == expected_normalized_frames
         assert result.notes == []
 
     def test_action_2_with_templates_and_special_case(self):
@@ -1008,7 +1165,7 @@ class TestSignatureGeneration:
         crash_data = {
             "os": "Windows NT",
             "crashing_thread": 0,
-            "threads": [frames_from_json_dump_with_templates_and_special_case],
+            "threads": [FRAMES_FROM_JSON_DUMP_WITH_TEMPLATES_AND_SPECIAL_CASE],
         }
         result = generator.Result()
 
@@ -1019,19 +1176,21 @@ class TestSignatureGeneration:
             result.signature
             == "<name omitted> | IPC::ParamTraits<mozilla::net::NetAddr>::Write"
         )
-        expected = (
-            "NtWaitForMultipleObjects | "
-            "<name omitted> | "
-            "IPC::ParamTraits<mozilla::net::NetAddr>::Write | "
-            "RealMsgWaitForMultipleObjectsEx | "
-            "MsgWaitForMultipleObjects | "
-            "F_1152915508__________________________________ | "
-            "F2166389______________________________________ | "
-            "F_917831355___________________________________ | "
-            "F1315696776________________________________ | "
-            "F_1428703866________________________________"
-        )
-        assert result.extra["proto_signature"] == expected
+        expected_normalized_frames = [
+            "NtWaitForMultipleObjects",
+            "<name omitted>",
+            "IPC::ParamTraits<mozilla::net::NetAddr>::Write",
+            "RealMsgWaitForMultipleObjectsEx",
+            "MsgWaitForMultipleObjects",
+            "F_1152915508__________________________________",
+            "F2166389______________________________________",
+            "F_917831355___________________________________",
+            "F1315696776________________________________",
+            "F_1428703866________________________________",
+        ]
+        expected_proto_signature = " | ".join(expected_normalized_frames)
+        assert result.extra["proto_signature"] == expected_proto_signature
+        assert result.extra["normalized_frames"] == expected_normalized_frames
         assert result.notes == []
 
     def test_action_3(self):
@@ -1043,10 +1202,11 @@ class TestSignatureGeneration:
         # the call to be tested
         assert sgr.action(crash_data, result) is True
 
-        assert result.signature == "EMPTY: no crashing thread identified"
+        assert result.signature == "EMPTY: no frame data available"
         assert "proto_signature" not in result.extra
+        assert "normalized_frames" not in result.extra
         assert result.notes == [
-            "SignatureGenerationRule: CSignatureTool: No signature could be created because we do not know which thread crashed"  # noqa
+            "SignatureGenerationRule: CSignatureTool: no frame data for crashing thread (0)"
         ]
 
     def test_lower_case_modules(self):
@@ -1076,11 +1236,50 @@ class TestSignatureGeneration:
         # the call to be tested
         assert sgr.action(crash_data, result) is True
         assert result.signature == "user2.dll"
-        expected = "@0x5e39bf21 | @0x5e39bf21 | @0x5e39bf21 | user2.dll@0x20869"
-        assert result.extra["proto_signature"] == expected
+        expected_normalized_frames = [
+            "@0x5e39bf21",
+            "@0x5e39bf21",
+            "@0x5e39bf21",
+            "user2.dll@0x20869",
+        ]
+        expected_proto_signature = " | ".join(expected_normalized_frames)
+        assert result.extra["proto_signature"] == expected_proto_signature
+        assert result.extra["normalized_frames"] == expected_normalized_frames
         assert result.notes == []
 
 
+class TestStackOverflowSignature:
+    def test_predicate_no_match(self):
+        result = generator.Result()
+        result.signature = "hello"
+        rule = rules.StackOverflowSignature()
+        assert rule.predicate({}, result) is False
+
+    @pytest.mark.parametrize(
+        "reason, expected",
+        [
+            ("FOO", False),
+            ("EXCEPTION_STACK_OVERFLOW", True),
+        ],
+    )
+    def test_predicate_reason(self, reason, expected):
+        crash_data = {"reason": reason}
+        result = generator.Result()
+        result.signature = "Text"
+        rule = rules.StackOverflowSignature()
+        assert rule.predicate(crash_data, result) is expected
+
+    def test_action_success(self):
+        crash_data = {"reason": "EXCEPTION_STACK_OVERFLOW"}
+        result = generator.Result()
+        result.signature = "hello"
+        rule = rules.StackOverflowSignature()
+        action_result = rule.action(crash_data, result)
+
+        assert action_result is True
+        assert result.signature == "stackoverflow | hello"
+
+
 class TestOOMSignature:
     def test_predicate_no_match(self):
         result = generator.Result()
@@ -1115,6 +1314,68 @@ class TestOOMSignature:
         rule = rules.OOMSignature()
         assert rule.predicate(crash_data, result) is True
 
+    @pytest.mark.parametrize(
+        "crashdata, expected",
+        [
+            ({}, False),
+            ({"crashing_thread": 5}, False),
+            ({"crashing_thread": 5, "threads": []}, False),
+            (
+                {
+                    "crashing_thread": 0,
+                    "threads": [
+                        {
+                            "last_error_value": "FOO",
+                        },
+                    ],
+                },
+                False,
+            ),
+            (
+                {
+                    "crashing_thread": 0,
+                    "threads": [
+                        {
+                            "last_error_value": "ERROR_COMMITMENT_LIMIT",
+                        },
+                    ],
+                },
+                True,
+            ),
+            (
+                {
+                    "crashing_thread": 0,
+                    "threads": [
+                        {
+                            "last_error_value": "ERROR_NOT_ENOUGH_MEMORY",
+                        },
+                    ],
+                },
+                True,
+            ),
+        ],
+    )
+    def test_predicate_error_commitment_limit(self, crashdata, expected):
+        result = generator.Result()
+        result.signature = "Text"
+        rule = rules.OOMSignature()
+        assert rule.predicate(crashdata, result) is expected
+
+    @pytest.mark.parametrize(
+        "reason, expected",
+        [
+            ("FOO", False),
+            ("STATUS_NO_MEMORY", True),
+            ("STATUS_FATAL_MEMORY_EXHAUSTION", True),
+        ],
+    )
+    def test_predicate_reason(self, reason, expected):
+        crash_data = {"reason": reason}
+        result = generator.Result()
+        result.signature = "Text"
+        rule = rules.OOMSignature()
+        assert rule.predicate(crash_data, result) is expected
+
     def test_action_success(self):
         crash_data = {}
         result = generator.Result()
@@ -1147,6 +1408,35 @@ class TestOOMSignature:
         assert result.signature == "OOM | large | hello"
 
 
+class TestBadHardware:
+    @pytest.mark.parametrize(
+        "reason, expected",
+        [
+            ("FOO", False),
+            ("EXCEPTION_IN_PAGE_ERROR_READ / STATUS_DEVICE_DATA_ERROR", True),
+            ("EXCEPTION_IN_PAGE_ERROR_EXEC / STATUS_DEVICE_DATA_ERROR", True),
+            ("EXCEPTION_IN_PAGE_ERROR_WRITE / STATUS_DEVICE_DATA_ERROR", True),
+        ],
+    )
+    def test_predicate_reason(self, reason, expected):
+        crash_data = {"reason": reason}
+        result = generator.Result()
+        result.signature = "Text"
+        rule = rules.BadHardware()
+        assert rule.predicate(crash_data, result) is expected
+
+    def test_action(self):
+        crash_data = {
+            "reason": "EXCEPTION_IN_PAGE_ERROR_READ / STATUS_DEVICE_DATA_ERROR",
+        }
+        result = generator.Result()
+        result.signature = "Text"
+        rule = rules.BadHardware()
+        action_result = rule.action(crash_data, result)
+        assert action_result is True
+        assert result.signature == "bad hardware | Text"
+
+
 class TestAbortSignature:
     def test_predicate(self):
         rule = rules.AbortSignature()
@@ -1323,14 +1613,6 @@ class TestStackwalkerErrorSignatureRule:
 
 
 class TestSignatureWatchDogRule:
-    def test_instantiation(self):
-        srwd = rules.SignatureRunWatchDog()
-
-        assert isinstance(srwd.c_signature_tool, rules.CSignatureTool)
-        assert isinstance(srwd.java_signature_tool, rules.JavaSignatureTool)
-
-        assert srwd._get_crashing_thread({}) == 0
-
     def test_predicate(self):
         srwd = rules.SignatureRunWatchDog()
 
@@ -1352,7 +1634,7 @@ class TestSignatureWatchDogRule:
         crash_data = {
             "os": "Windows NT",
             "crashing_thread": 0,
-            "threads": [frames_from_json_dump],
+            "threads": [FRAMES_FROM_JSON_DUMP],
         }
         result = generator.Result()
         result.signature = "foo::bar"
@@ -1369,42 +1651,6 @@ class TestSignatureWatchDogRule:
         assert result.notes == []
 
 
-class TestSignatureJitCategory:
-    def test_predicate_no_match(self):
-        rule = rules.SignatureJitCategory()
-
-        crash_data = {}
-        result = {"signature": "", "notes": []}
-        assert rule.predicate(crash_data, result) is False
-
-        crash_data = {"jit_category": ""}
-        result = {"signature": "", "notes": []}
-
-        assert rule.predicate(crash_data, result) is False
-
-    def test_predicate(self):
-        rule = rules.SignatureJitCategory()
-
-        crash_data = {"jit_category": "JIT Crash"}
-        result = {"signature": "", "notes": []}
-
-        assert rule.predicate(crash_data, result) is True
-
-    def test_action_success(self):
-        rule = rules.SignatureJitCategory()
-
-        crash_data = {"jit_category": "JIT Crash"}
-        result = generator.Result()
-        result.signature = "foo::bar"
-
-        action_result = rule.action(crash_data, result)
-        assert action_result is True
-        assert result.signature == "jit | JIT Crash"
-        assert result.notes == [
-            'SignatureJitCategory: Signature replaced with a JIT Crash Category, was: "foo::bar"'
-        ]
-
-
 class TestSignatureIPCChannelError:
     def test_predicate_no_match(self):
         rule = rules.SignatureIPCChannelError()
@@ -1442,7 +1688,7 @@ class TestSignatureIPCChannelError:
         ]
 
         # Now test with a browser crash.
-        crash_data["additional_minidumps"] = "browser"
+        crash_data["additional_minidumps"] = ["upload_file_minidump_browser"]
         result = generator.Result()
         result.signature = original_signature
 
@@ -1454,25 +1700,6 @@ class TestSignatureIPCChannelError:
             "SignatureIPCChannelError: IPC Channel Error stomped on signature"
         ]
 
-    def test_shutdownkill(self):
-        rule = rules.SignatureIPCChannelError()
-
-        original_signature = "foo::bar"
-        crash_data = {
-            "ipc_channel_error": "ShutDownKill",
-            "additional_minidumps": "browser",
-        }
-        result = generator.Result()
-        result.signature = original_signature
-
-        action_result = rule.action(crash_data, result)
-        assert action_result is True
-
-        assert result.signature == "IPCError-browser | ShutDownKill | {}".format(
-            original_signature
-        )
-        assert result.notes == ["SignatureIPCChannelError: IPC Channel Error prepended"]
-
 
 class TestSignatureShutdownTimeout:
     def test_predicate_no_match(self):
@@ -1520,11 +1747,11 @@ class TestSignatureShutdownTimeout:
 
         action_result = rule.action(crash_data, result)
         assert action_result is True
-        assert result.signature == "AsyncShutdownTimeout | UNKNOWN"
+        assert result.signature == "AsyncShutdownTimeout | (unknown) | (none)"
 
         assert result.notes == [
-            "SignatureShutdownTimeout: Error parsing AsyncShutdownTimeout: 'phase'",
-            'SignatureShutdownTimeout: Signature replaced with a Shutdown Timeout signature, was: "foo"',  # noqa
+            "SignatureShutdownTimeout: Signature replaced with a Shutdown "
+            + 'Timeout signature, was: "foo"',
         ]
 
     def test_action_success(self):
@@ -1542,7 +1769,8 @@ class TestSignatureShutdownTimeout:
         assert action_result is True
         assert result.signature == "AsyncShutdownTimeout | beginning | A,B"
         assert result.notes == [
-            'SignatureShutdownTimeout: Signature replaced with a Shutdown Timeout signature, was: "foo"'  # noqa
+            "SignatureShutdownTimeout: Signature replaced with a Shutdown "
+            + 'Timeout signature, was: "foo"',
         ]
 
     def test_action_success_string_conditions(self):
@@ -1560,7 +1788,8 @@ class TestSignatureShutdownTimeout:
         assert action_result is True
         assert result.signature == "AsyncShutdownTimeout | beginning | A,B,C"
         assert result.notes == [
-            'SignatureShutdownTimeout: Signature replaced with a Shutdown Timeout signature, was: "foo"'  # noqa
+            "SignatureShutdownTimeout: Signature replaced with a Shutdown "
+            + 'Timeout signature, was: "foo"',
         ]
 
     def test_action_success_empty_conditions_key(self):
@@ -1578,7 +1807,23 @@ class TestSignatureShutdownTimeout:
         assert action_result is True
         assert result.signature == "AsyncShutdownTimeout | beginning | (none)"
         assert result.notes == [
-            'SignatureShutdownTimeout: Signature replaced with a Shutdown Timeout signature, was: "foo"'  # noqa
+            "SignatureShutdownTimeout: Signature replaced with a Shutdown "
+            + 'Timeout signature, was: "foo"',
+        ]
+
+    def test_action_unsupported_structure(self):
+        rule = rules.SignatureShutdownTimeout()
+
+        crash_data = {"async_shutdown_timeout": json.dumps({})}
+        result = generator.Result()
+        result.signature = "foo"
+
+        action_result = rule.action(crash_data, result)
+        assert action_result is True
+        assert result.signature == "AsyncShutdownTimeout | (unknown) | (none)"
+        assert result.notes == [
+            "SignatureShutdownTimeout: Signature replaced with a Shutdown "
+            + 'Timeout signature, was: "foo"',
         ]
 
 
